@@ -1,11 +1,11 @@
 # Copyright 2011 Omniscale GmbH & Co. KG
-# 
+#
 # Licensed under the Apache License, Version 2.0 (the "License");
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
-# 
+#
 #     http://www.apache.org/licenses/LICENSE-2.0
-# 
+#
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS,
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
@@ -32,14 +32,14 @@ class MMapReader(object):
         self.m = m
         self.m.seek(0)
         self.size = size
-    
+
     def read(self, size=None):
         if size is None:
             size = self.size - self.m.tell()
         else:
             size = min(self.size - self.m.tell(), size)
         return self.m.read(size)
-        
+
     def readline(self):
         cur_pos = self.m.tell()
         if cur_pos >= self.size:
@@ -47,17 +47,17 @@ class MMapReader(object):
         nl_pos = self.m.find('\n')
         self.m.seek(cur_pos)
         return self.m.read(nl_pos-cur_pos)
-    
+
     def seek(self, n):
         self.m.seek(n)
 
 class XMLParserProcess(XMLParser, multiprocessing.Process):
     def __init__(self, mmap_pool, mmap_queue, *args, **kw):
+        multiprocessing.Process.__init__(self)
+        XMLParser.__init__(self, *args, **kw)
+        self.daemon = True
         self.mmap_pool = mmap_pool
         self.mmap_queue = mmap_queue
-        XMLParser.__init__(self, *args, **kw)
-        multiprocessing.Process.__init__(self)
-        self.daemon = True
 
     def run(self):
         setproctitle('imposm xml parser')
@@ -77,7 +77,7 @@ class XMLMultiProcParser(object):
     nodes_tag_filter = None
     ways_tag_filter = None
     relations_tag_filter = None
-    
+
     def __init__(self, pool_size, nodes_queue=None, ways_queue=None,
         relations_queue=None, coords_queue=None, marshal_elem_data=False):
         self.pool_size = pool_size
@@ -90,10 +90,10 @@ class XMLMultiProcParser(object):
         self.mmap_pool = MMapPool(pool_size*8, xml_chunk_size*8)
         self.mmap_queue = multiprocessing.JoinableQueue(8)
         self.marshal_elem_data = marshal_elem_data
-        
+
     def parse(self, stream):
         assert not self.pool
-        
+
         for _ in xrange(self.pool_size):
             proc = XMLParserProcess(self.mmap_pool, self.mmap_queue, nodes_callback=self.nodes_callback,
                 coords_callback=self.coords_callback, ways_callback=self.ways_callback,
@@ -105,16 +105,16 @@ class XMLMultiProcParser(object):
             )
             self.pool.append(proc)
             proc.start()
-        
+
         chunker = XMLChunker(stream, self.mmap_pool, xml_chunk_size=READ_SIZE)
         chunker.read(self.mmap_queue, coords_callback=self.coords_callback)
-        
+
         self.mmap_queue.join()
         for proc in self.pool:
             self.mmap_queue.put((None, None))
         for proc in self.pool:
             proc.join()
-        
+
 
 class MMapPool(object):
     """
@@ -127,11 +127,11 @@ class MMapPool(object):
         self.pool = [mmap.mmap(-1, mmap_size) for _ in range(n)]
         self.free_mmaps = set(range(n))
         self.free_queue = multiprocessing.JoinableQueue()
-        
+
     def new(self):
         """
         Return a free mmap file.
-        
+
         :returns: index, mmap file
         """
         if not self.free_mmaps:
@@ -146,7 +146,7 @@ class MMapPool(object):
                 break
         mmap_idx = self.free_mmaps.pop()
         return mmap_idx, self.pool[mmap_idx]
-    
+
     def join(self):
         while len(self.free_mmaps) < self.n:
             self.free_mmaps.add(self.free_queue.get())
@@ -167,10 +167,10 @@ class MMapPool(object):
 class XMLChunker(object):
     """
     Reads and chunks OSM XML file.
-    
+
     Reads OSM XML from `stream` and writes chunks of it into mmap files from
     the `mmap_pool`.
-    
+
     :params xml_chunk_size: chunk XML after this many bytes
     """
     def __init__(self, stream, mmap_pool, xml_chunk_size):
@@ -200,13 +200,14 @@ class XMLChunker(object):
 
     def read(self, mmaps_queue, coords_callback=None):
         """
-        Read and chunk all 
+        Read and chunk all
         """
         coord_node_match = None
         xml_nodes = self._new_xml_outstream()
         coords = []
         coord_node_re_match = re.compile(r'^\s*<node id="(\d+)" .*lat="([-0-9.]+)" '
                                           'lon="([-0-9.]+)".*/>').match
+        node_re_match = re.compile(r'^\s*<node .*/>').match
         xml_nodes.write(self._last_line)
         split = False
         line = ''
@@ -226,7 +227,7 @@ class XMLChunker(object):
             if split:
                 if (line.lstrip().startswith('</')
                     or (coords_callback and coord_node_match)
-                    or (not coords_callback and coord_node_re_match(line))):
+                    or (not coords_callback and node_re_match(line))):
                     mmaps_queue.put(self._finished_xml_outstream(line, xml_nodes))
                     xml_nodes = self._new_xml_outstream()
                     split = False
@@ -234,7 +235,11 @@ class XMLChunker(object):
                 split = True
         if coords_callback:
             coords_callback(coords)
-        
+
+        # we are at the end of the stream and assume we wrote the end tag
+        # to xml_nodes. we set line to closing tag here to avoid additional
+        # end tag in case the last line(s) is blank
+        line = '</osm'
         mmaps_queue.put(self._finished_xml_outstream(line, xml_nodes))
 
 if __name__ == '__main__':
@@ -253,13 +258,13 @@ if __name__ == '__main__':
                 queue.task_done()
             print type, count
         return count
-    
-    
+
+
     nodes_queue = multiprocessing.JoinableQueue(128)
     ways_queue = multiprocessing.JoinableQueue(128)
     relations_queue = multiprocessing.JoinableQueue(128)
 
-    
+
     procs = [
         multiprocessing.Process(target=count_proc('nodes', nodes_queue)),
         multiprocessing.Process(target=count_proc('ways', ways_queue)),
@@ -267,14 +272,14 @@ if __name__ == '__main__':
     ]
     for proc in procs:
         proc.start()
-    
+
     parser = XMLMultiProcParser(open(sys.argv[1]), 2, nodes_queue=nodes_queue,
         ways_queue=ways_queue, relations_queue=relations_queue)
     parser.start()
-    
+
     nodes_queue.put(None)
     nodes_queue.join()
-    
+
     ways_queue.put(None)
     ways_queue.join()
     relations_queue.put(None)
